'use strict';

/**
 *
 * jQuery Deferred 1.7
 * Some code (c) 2005, 2013 jQuery Foundation, Inc. and other contributors
 *
 */


// String to Object flags format cache
const flagsCache = {};
const class2type = {};
// Save a reference to some core methods
const toString = Object.prototype.toString;
const hasOwn = Object.prototype.hasOwnProperty;

// Convert String-formatted flags into Object-formatted ones and store in cache
function createFlags(flags) {
    const object = flagsCache[flags] = {};

    let i,
        length;

    flags = flags.split(/\s+/);

    for (i = 0, length = flags.length; i < length; i++) {
        object[flags[i]] = true;
    }

    return object;
}

const jQuery = {

    // See test/unit/core.js for details concerning isFunction.
    // Since version 1.3, DOM methods and functions like alert
    // aren't supported. They return false on IE (#2968).
    isFunction: function(obj) {
        return jQuery.type(obj) === 'function';
    },

    isArray: Array.isArray || function(obj) {
        return jQuery.type(obj) === 'array';
    },

    // A crude way of determining if an object is a window
    isWindow: function(obj) {
        return obj && typeof obj === 'object' && 'setInterval' in obj;
    },

    isNumeric: function(obj) {
        return !isNaN(parseFloat(obj)) && isFinite(obj);
    },

    type: function(obj) {
        return obj == null ?
            String(obj) :
            class2type[toString.call(obj)] || 'object';
    },

    isPlainObject: function(obj) {
        // Must be an Object.
        // Because of IE, we also have to check the presence of the constructor property.
        // Make sure that DOM nodes and window objects don't pass through, as well
        if (!obj || jQuery.type(obj) !== 'object' || obj.nodeType || jQuery.isWindow(obj)) {
            return false;
        }

        try {
            // Not own constructor property must be Object
            if (obj.constructor &&
                !hasOwn.call(obj, 'constructor') &&
                !hasOwn.call(obj.constructor.prototype, 'isPrototypeOf')) {
                return false;
            }
        } catch (e) {
            // IE8,9 Will throw exceptions on certain host objects #9897
            return false;
        }

        // Own properties are enumerated firstly, so to speed up,
        // if last one is own, then all properties are own.

        let key,
            a;

        for (a in obj) {
            key = a;
        }

        return key === undefined || hasOwn.call(obj, key);
    },

    isEmptyObject: function(obj) {
        for (const name in obj) {
            return false;
        }
        return true;
    },

    error: function(msg) {
        throw new Error(msg);
    },

    // args is for internal usage only
    each: function(object, callback, args) {
        let name,
            i = 0;

        const length = object.length;
        const isObj = length === undefined || jQuery.isFunction(object);

        if (args) {
            if (isObj) {
                for (name in object) {
                    if (callback.apply(object[name], args) === false) {
                        break;
                    }
                }
            } else {
                for (; i < length;) {
                    if (callback.apply(object[i++], args) === false) {
                        break;
                    }
                }
            }

            // A special, fast, case for the most common use of each
        } else {
            if (isObj) {
                for (name in object) {
                    if (callback.call(object[name], name, object[name]) === false) {
                        break;
                    }
                }
            } else {
                for (; i < length;) {
                    if (callback.call(object[i], i, object[i++]) === false) {
                        break;
                    }
                }
            }
        }

        return object;
    }

};

// Populate the class2type map
jQuery.each('Boolean Number String Function Array Date RegExp Object'.split(' '), function(i, name) {
    class2type['[object ' + name + ']'] = name.toLowerCase();
});


jQuery.extend = function() {
    let options,
        name,
        src,
        copy,
        copyIsArray,
        clone,
        target = arguments[0] || {},
        i = 1,
        deep = false;

    const length = arguments.length;

    // Handle a deep copy situation
    if (typeof target === 'boolean') {
        deep = target;
        target = arguments[1] || {};
        // skip the boolean and the target
        i = 2;
    }

    // Handle case when target is a string or something (possible in deep copy)
    if (typeof target !== 'object' && !jQuery.isFunction(target)) {
        target = {};
    }

    // extend jQuery itself if only one argument is passed
    if (length === i) {
        target = this;
        --i;
    }

    for (; i < length; i++) {
        // Only deal with non-null/undefined values
        if ((options = arguments[i]) != null) {
            // Extend the base object
            for (name in options) {
                src = target[name];
                copy = options[name];

                // Prevent never-ending loop
                if (target === copy) {
                    continue;
                }

                // Recurse if we're merging plain objects or arrays
                if (deep && copy && (jQuery.isPlainObject(copy) || (copyIsArray = jQuery.isArray(copy)))) {
                    if (copyIsArray) {
                        copyIsArray = false;
                        clone = src && jQuery.isArray(src) ? src : [];

                    } else {
                        clone = src && jQuery.isPlainObject(src) ? src : {};
                    }

                    // Never move original objects, clone them
                    target[name] = jQuery.extend(deep, clone, copy);

                    // Don't bring in undefined values
                } else if (copy !== undefined) {
                    target[name] = copy;
                }
            }
        }
    }

    // Return the modified object
    return target;
};

/*
 * Create a callback list using the following parameters:
 *
 *    flags:    an optional list of space-separated flags that will change how
 *            the callback list behaves
 *
 * By default a callback list will act like an event callback list and can be
 * "fired" multiple times.
 *
 * Possible flags:
 *
 *    once:            will ensure the callback list can only be fired once (like a Deferred)
 *
 *    memory:            will keep track of previous values and will call any callback added
 *                    after the list has been fired right away with the latest "memorized"
 *                    values (like a Deferred)
 *
 *    unique:            will ensure a callback can only be added once (no duplicate in the list)
 *
 *    stopOnFalse:    interrupt callings when a callback returns false
 *
 */
jQuery.Callbacks = function(flags) {

    // Convert flags from String-formatted to Object-formatted
    // (we check in cache first)
    flags = flags ? (flagsCache[flags] || createFlags(flags)) : {};

    let // Actual callback list
        list = [],
        // Stack of fire calls for repeatable lists
        stack = [],
        // Last fire value (for non-forgettable lists)
        memory,
        // Flag to know if list is currently firing
        firing,
        // First callback to fire (used internally by add and fireWith)
        firingStart,
        // End of the loop when firing
        firingLength,
        // Index of currently firing callback (modified by remove if needed)
        firingIndex;

        // Add one or several callbacks to the list
    const add = function(args) {
        let i,
            length,
            elem,
            type;

        for (i = 0, length = args.length; i < length; i++) {
            elem = args[i];
            type = jQuery.type(elem);
            if (type === 'array') {
                // Inspect recursively
                add(elem);
            } else if (type === 'function') {
                // Add if not in unique mode and callback is not in
                if (!flags.unique || !self.has(elem)) {
                    elem.__domain = process.domain;
                    list.push(elem);
                }
            }
        }
    };

        // Fire callbacks
    const fire = function(context, args) {

        let fn,
            domain,
            ret;

        args = args || [];
        memory = !flags.memory || [context, args];
        firing = true;
        // firingIndex = firingStart || 0;
        // firingStart = 0;
        firingLength = list && list.length || 0;
        for (firingIndex = firingStart || 0, firingStart = 0; firingIndex < firingLength; firingIndex++) {

            fn = list[firingIndex];
            domain = fn.__domain;
            fn.__domain = undefined;

            // restore domain if needed
            if (domain && domain !== process.domain) {
                domain.run(function() {                 // eslint-disable-line no-loop-func
                    ret = fn.apply(context, args);
                });
            } else {
                ret = fn.apply(context, args);
            }

            if (ret === false && flags.stopOnFalse) {
                memory = true; // Mark as halted
                break;
            }
        }
        firing = false;
        if (list) {
            if (!flags.once) {
                if (stack && stack.length) {
                    memory = stack.shift();
                    self.fireWith(memory[0], memory[1]);
                }
            } else
            if (memory === true) {
                self.disable();
            } else {
                list = [];
            }
        }
    };

        // Actual Callbacks object
    const self = {
        // Add a callback or a collection of callbacks to the list
        add: function() {
            if (list) {
                const length = list.length;
                add(arguments);
                // Do we need to add the callbacks to the
                // current firing batch?
                if (firing) {
                    firingLength = list.length;
                    // With memory, if we're not firing then
                    // we should call right away, unless previous
                    // firing was halted (stopOnFalse)
                } else if (memory && memory !== true) {
                    firingStart = length;
                    fire(memory[0], memory[1]);
                }
            }
            return this;
        },
        // Remove a callback from the list
        remove: function() {
            if (list) {
                const args = arguments;

                const argLength = args.length;

                let argIndex = 0;

                for (; argIndex < argLength; argIndex++) {
                    for (let i = 0; i < list.length; i++) {
                        if (args[argIndex] === list[i]) {
                            // Handle firingIndex and firingLength
                            if (firing) {
                                if (i <= firingLength) {
                                    firingLength--;
                                    if (i <= firingIndex) {
                                        firingIndex--;
                                    }
                                }
                            }
                            // Remove the element
                            list.splice(i--, 1);
                            // If we have some unicity property then
                            // we only need to do this once
                            if (flags.unique) {
                                break;
                            }
                        }
                    }
                }
            }
            return this;
        },
        // Control if a given callback is in the list
        has: function(fn) {
            if (list) {
                const length = list.length;

                let i = 0;
                for (; i < length; i++) {
                    if (fn === list[i]) {
                        return true;
                    }
                }
            }
            return false;
        },
        // Remove all callbacks from the list
        empty: function() {
            list = [];
            return this;
        },
        // Have the list do nothing anymore
        disable: function() {
            list = stack = memory = undefined;
            return this;
        },
        // Is it disabled?
        disabled: function() {
            return !list;
        },
        // Lock the list in its current state
        lock: function() {
            stack = undefined;
            if (!memory || memory === true) {
                self.disable();
            }
            return this;
        },
        // Is it locked?
        locked: function() {
            return !stack;
        },
        // Call all callbacks with the given context and arguments
        fireWith: function(context, args) {
            if (stack) {
                if (firing) {
                    if (!flags.once) {
                        stack.push([context, args]);
                    }
                } else if (!(flags.once && memory)) {
                    fire(context, args);
                }
            }
            return this;
        },
        // Call all the callbacks with the given arguments
        fire: function() {
            self.fireWith(this, arguments);
            return this;
        },
        // To know if the callbacks have already been called at least once
        fired: function() {
            return !!memory;
        }
    };

    return self;
};


const // Static reference to slice
    sliceDeferred = [].slice;

jQuery.extend({

    Deferred: function(func) {
        const doneList = jQuery.Callbacks('once memory');
        const failList = jQuery.Callbacks('once memory');
        const progressList = jQuery.Callbacks('memory');

        let state = 'pending';

        const lists = {
            resolve: doneList,
            reject: failList,
            notify: progressList
        };

        const promise = {
            done: doneList.add,
            fail: failList.add,
            progress: progressList.add,

            state: function() {
                return state;
            },

            // Deprecated
            isResolved: doneList.fired,
            isRejected: failList.fired,

            then: function(doneCallbacks, failCallbacks, progressCallbacks) {
                deferred.done(doneCallbacks).fail(failCallbacks).progress(progressCallbacks);
                return this;
            },
            always: function() {
                deferred.done.apply(deferred, arguments).fail.apply(deferred, arguments);
                return this;
            },

            toES6Promise: function() {
                const self = this;

                return new Promise(function (resolve, reject) {

                    self.done(function(d) {
                        resolve(d);
                    });

                    self.fail(function(d) {
                        reject(d);
                    });
                });
            },
            pipe: function(fnDone, fnFail, fnProgress) {
                return jQuery.Deferred(function(newDefer) {
                    jQuery.each({
                        done: [fnDone, 'resolve'],
                        fail: [fnFail, 'reject'],
                        progress: [fnProgress, 'notify']
                    }, function(handler, data) {
                        const fn = data[0];
                        const action = data[1];

                        let returned;
                        if (jQuery.isFunction(fn)) {
                            deferred[handler](function() {
                                returned = fn.apply(this, arguments);
                                if (returned && jQuery.isFunction(returned.promise)) {
                                    returned.promise().then(newDefer.resolve, newDefer.reject, newDefer.notify);
                                } else {
                                    newDefer[action + 'With'](this === deferred ? newDefer : this, [returned]);
                                }
                            });
                        } else {
                            deferred[handler](newDefer[action]);
                        }
                    });
                }).promise();
            },
            // Get a promise for this deferred
            // If obj is provided, the promise aspect is added to the object
            promise: function(obj) {
                if (obj == null) {
                    obj = promise;
                } else {
                    for (const key in promise) {
                        obj[key] = promise[key];
                    }
                }
                return obj;
            }
        };

        const deferred = promise.promise({});

        let key;

        for (key in lists) {
            deferred[key] = lists[key].fire;
            deferred[key + 'With'] = lists[key].fireWith;
        }

        // Handle state
        deferred.done(function() {
            state = 'resolved';
        }, failList.disable, progressList.lock).fail(function() {
            state = 'rejected';
        }, doneList.disable, progressList.lock);

        // Call given func if any
        if (func) {
            func.call(deferred, deferred);
        }

        // All done!
        return deferred;
    },

    // Deferred helper
    when: function(firstParam) {
        const args = sliceDeferred.call(arguments, 0);
        const length = args.length;
        const pValues = new Array(length);
        const deferred = length <= 1 && firstParam && jQuery.isFunction(firstParam.promise) ?
            firstParam :
            jQuery.Deferred();
        const promise = deferred.promise();

        let i = 0,
            count = length;

        function resolveFunc(i) {
            return function(value) {
                args[i] = arguments.length > 1 ? sliceDeferred.call(arguments, 0) : value;
                if (!(--count)) {
                    deferred.resolveWith(deferred, args);
                }
            };
        }
        function progressFunc(i) {
            return function(value) {
                pValues[i] = arguments.length > 1 ? sliceDeferred.call(arguments, 0) : value;
                deferred.notifyWith(promise, pValues);
            };
        }
        if (length > 1) {
            for (; i < length; i++) {
                if (args[i] && args[i].promise && jQuery.isFunction(args[i].promise)) {
                    args[i].promise().then(resolveFunc(i), deferred.reject, progressFunc(i));
                } else {
                    --count;
                }
            }
            if (!count) {
                deferred.resolveWith(deferred, args);
            }
        } else if (deferred !== firstParam) {
            deferred.resolveWith(deferred, length ? [firstParam] : []);
        }
        return promise;
    }
});


module.exports = {
    create: function() {
        return jQuery.Deferred();
    },
    when: jQuery.when,
    extend: jQuery.extend
};

